from numba import njit, f8, i8
from numba.types import ListType
from numba.typed import List
from cre.conditions import OR, AND, literal_ctor, literal_not
from cre.memset import MemSet
from cre.context import cre_context
from cre.obj import CREObjType
from cre.fact import define_fact
from cre.var import Var
from time import time_ns
from cre.utils import  _raw_ptr_from_struct, _cast_structref

import  cre.dynamic_exec



# BOOP = define_fact("BOOP",{"A": "string", "B" : "number"})

#TODO: would be nice to have functionality like x:=Var(BOOP) => Var(BOOP,'x')
def _test_auto_aliasing():
    pass


def print_str_diff(a,b):
    import difflib
    for i,s in enumerate(difflib.ndiff(a, b)):
        if s[0]==' ': continue
        elif s[0]=='-':
            print(u'Delete "{}" from position {}'.format(s[-1],i))
        elif s[0]=='+':
            print(u'Add "{}" to position {}'.format(s[-1],i))    

def test_var():
    with cre_context("test_var"):
        BOOP = define_fact("BOOP",{"A": "string", "B" : "number"})

        @njit(cache=False)
        def var_to_str(v):
            return str(v)

        Q = Var(BOOP,'Q')

        print(Q.copy())

        v = Q.A
        print(v)
        assert var_to_str(v) == "Q.A"
        assert str(v) == "Q.A"

        Z = Q.with_alias("Z")
        print(Z)
        assert var_to_str(Z) == "Z"
        assert str(Z) == "Z"

        x = v.with_alias("Z")
        print(x)
        assert var_to_str(x) == "Z.A"
        assert str(x) == "Z.A"

        l1 = Var(BOOP)
        l2 = Var(BOOP)
        assert l1.base_ptr == l1.B.base_ptr 
        assert l2.base_ptr == l2.B.base_ptr 
        



# from cre.conditions import conds_repr, conditions_repr
# @njit(cache=True)
def test_build_conditions():
    with cre_context("test_build_conditions"):
        BOOP = define_fact("BOOP",{"A": "string", "B" : "number"})

        # l1, l2 = Var(BOOP,"l1"), Var(BOOP,"l2")
        # r1, r2 = Var(BOOP,"r1"), Var(BOOP,"r2")

        # (l1:=Var(BOOP, 'l1')) & l1 
        # a = (  & l1 )

        
        c = \
        AND(l1:=Var(BOOP), l1.B < 1, l1.B > 7,
            r1:=Var(BOOP), l1.B < r1.B,
            r2:=Var(BOOP), l1.B < r2.B,
            l2:=Var(BOOP)
        )


        print(c)

        c = \
        OR((
        (l1:=Var(BOOP)) & (l1.B < 1) & (l1.B > 7) &
        (r1:=Var(BOOP)) & (l1.B < r1.B) &
        (r2:=Var(BOOP)) & (l1.B < r2.B) &
        (l2:=Var(BOOP))
        ),(
         l1 &
         r1 &
         r2 &
         l2 & (l2.B < 1) & (l2.B > 7) & (l2.B < r1.B) & (l2.B < r2.B))
        )
        print(c)
        # raise ValueError()



        c1 = l1.B < 1
        c2 = l1.B < l2.B

        print(c1)
        print(c2)

        assert str(c1) == "l1.B < 1"
        assert str(c2) == "l1.B < l2.B"



        c12 = (l1.B < 1) & (l1.B > 7)

        print(c12)
        assert str(c12) == "AND(l1:=Var(BOOP), l1.B < 1, l1.B > 7)"

        ### LT + AND/OR ###

        c3 = OR(
                (l1.B < 1) & (l1.B > 7) & (l1.B < r1.B) & (l1.B < r2.B),
                (l2.B < 1) & (l2.B > 7) & (l2.B < r1.B) & (l2.B < r2.B)
               )

        c3_str = \
'''OR(
    AND(l1:=Var(BOOP), l1.B < 1, l1.B > 7,
        r1:=Var(BOOP), l1.B < r1.B,
        r2:=Var(BOOP), l1.B < r2.B,
        l2:=Var(BOOP)
    ),
    AND(l1,
        r1,
        r2,
        l2, l2.B < 1, l2.B > 7, l2.B < r1.B, l2.B < r2.B
    )
)'''    
        print(c3)
        assert str(c3) == c3_str

        c3_repr = \
'''OR(
 AND(l1.B < 1, l1.B > 7, l1.B < r1.B, l1.B < r2.B),
 AND(l2.B < 1, l2.B > 7, l2.B < r1.B, l2.B < r2.B)
)'''
        print(repr(c3))
        assert repr(c3) == c3_repr

        ### NOT ###
        n3 = ~c3
        # print(n3)
        print(repr(n3))
        nc3_repr = \
'''OR(
 AND(~(l1.B < 1), ~(l2.B < 1)),
 AND(~(l1.B < 1), ~(l2.B > 7)),
 AND(~(l1.B < 1), ~(l2.B < r1.B)),
 AND(~(l1.B < 1), ~(l2.B < r2.B)),
 AND(~(l1.B > 7), ~(l2.B < 1)),
 AND(~(l1.B > 7), ~(l2.B > 7)),
 AND(~(l1.B > 7), ~(l2.B < r1.B)),
 AND(~(l1.B > 7), ~(l2.B < r2.B)),
 AND(~(l1.B < r1.B), ~(l2.B < 1)),
 AND(~(l1.B < r1.B), ~(l2.B > 7)),
 AND(~(l1.B < r1.B), ~(l2.B < r1.B)),
 AND(~(l1.B < r1.B), ~(l2.B < r2.B)),
 AND(~(l1.B < r2.B), ~(l2.B < 1)),
 AND(~(l1.B < r2.B), ~(l2.B > 7)),
 AND(~(l1.B < r2.B), ~(l2.B < r1.B)),
 AND(~(l1.B < r2.B), ~(l2.B < r2.B))
)'''

        # print_str_diff(repr(n3), nc3_repr)  
        # print(list(difflib.ndiff(repr(n3), nc3_repr)))
        assert repr(n3) == nc3_repr

        ### EQ / NEQ ###

        c4 = (l1.B == 5) & (l1.B == 5) & (l1.B == l2.B) & (l1.B != l2.B)

        c4_str = \
'''AND(l1:=Var(BOOP), l1.B == 5, l1.B == 5,
    l2:=Var(BOOP), l1.B == l2.B, ~(l1.B == l2.B)
)'''
        # print()
        print(str(c4))
        # print_str_diff(str(c4), c4_str)
        # print(repr(c4))
        assert str(c4) == c4_str    

        c4_repr = \
'''AND(l1.B == 5, l1.B == 5, l1.B == l2.B, ~(l1.B == l2.B))'''

        print(repr(c4))
        assert repr(c4) == c4_repr    

        nc4 = ~c4
        # print(nc4)
        print(repr(nc4))

        nc4_repr = \
'''OR(
 AND(~(l1.B == 5)),
 AND(~(l1.B == 5)),
 AND(~(l1.B == l2.B)),
 AND(l1.B == l2.B)
)'''
        assert repr(nc4) == nc4_repr

        ### AND / OR btw DNFS ### 
        c3_and_c4_repr = \
'''OR(
 AND(l1.B < 1, l1.B > 7, l1.B < r1.B, l1.B < r2.B, l1.B == 5, l1.B == 5, l1.B == l2.B, ~(l1.B == l2.B)),
 AND(l2.B < 1, l2.B > 7, l2.B < r1.B, l2.B < r2.B, l1.B == 5, l1.B == 5, l1.B == l2.B, ~(l1.B == l2.B))
)'''        
        # print()
        print(repr(c3 & c4))
        assert repr(c3 & c4) == c3_and_c4_repr

        c3_or_c4_repr = \
'''OR(
 AND(l1.B < 1, l1.B > 7, l1.B < r1.B, l1.B < r2.B),
 AND(l2.B < 1, l2.B > 7, l2.B < r1.B, l2.B < r2.B),
 AND(l1.B == 5, l1.B == 5, l1.B == l2.B, ~(l1.B == l2.B))
)'''
        print(repr(c3 | c4))
        assert repr(c3 | c4) == c3_or_c4_repr


list_i8 = ListType(i8)
list_list_i8 = ListType(ListType(i8))

@njit(cache=True)
def get_init_cond_sizes(conds):
    alpha_sizes = List.empty_list(list_i8)
    beta_sizes = List.empty_list(list_i8)
    for alpha_conjucts, beta_conjucts, beta_inds in conds.distr_dnf:
        alpha_sizes.append(List([len(conjunct) for conjunct in alpha_conjucts]))
        beta_sizes.append(List([len(conjunct) for conjunct in beta_conjucts]))

    return alpha_sizes, beta_sizes


@njit(cache=True)
def var_get_ptr(var):
    return _raw_ptr_from_struct(var)

@njit(cache=True)
def cond_get_vars(cond):
    return cond.vars


@njit(cache=True)
def get_pointer(st):
    return _raw_ptr_from_struct(st)

def _test_link():
    '''TODO: REWRITE'''
    print("START TEST LINK")
    with cre_context("test_link") as context:
        BOOP = define_fact("BOOP",{"A": "string", "B" : "number"})
        
        l1, l2 = Var(BOOP,"l1"), Var(BOOP,"l2")
        r1, r2 = Var(BOOP,"r1"), Var(BOOP,"r2")

        c = (l1.B < 1) & (l1.B > 7) & (l2.B < r1.B) & (r2.B < l1.B) |\
             (l2.B < 1) & (l2.B > 7) & (l2.B < r1.B) & (r1.B < r2.B) |\
             (l2.B < 1) & (l2.B > 7) & (l2.B < r1.B) & (l1.B < l2.B)

        print(c)
        ms = MemSet()
        cl = get_linked_conditions_instance(c, ms)

        assert get_pointer(cl) == get_pointer(c)

        cl = get_linked_conditions_instance(c, ms, copy=True)

        assert get_pointer(cl) != get_pointer(c)

        

# def test_unconditioned():
#     with cre_context("test_unconditioned") as context:
#         BOOP = define_fact("BOOP",{"A": "string", "B" : "number"})
#         l1, l2 = Var(BOOP,"l1"), Var(BOOP,"l2")

#         # c = var_and(l1,l2)

#         print(conditions_repr(l1 & l2,"c"))
#         print(conditions_repr((l1.B < 1) & l2, "c"))
#         print(conditions_repr(l1 & (l2.B > 1), "c"))
    
def test_multiple_deref():
    with cre_context("test_ref_matching"):
        TestLL = define_fact("TestLL", {"name": "string", "B" :'number', "nxt" : "TestLL"})

        v1 = Var(TestLL,'v1')
        v2 = Var(TestLL,'v2')
        c = v1.nxt.nxt == v2.nxt.nxt
        print(c)

def _test_existential_not():
    with cre_context("test_existential_not"):
        l1, l2 = Var(BOOP,"l1"), Var(BOOP,"l2")
        # print(l1.B)
        # print(~l1)
        # print(NOT(l1).B)
        # print(NOT(l1.B))
        a = (l1.B < 1)
        # print(repr(a))
        c = a & (l2.B > 1)
        c_n = NOT(c)
        # print("c.vars",c.vars)
        # print("c_n.vars", c_n.vars)
        print(repr(c))
        print(repr(c_n))
        assert repr(c) == 'l1, l2 = Var(BOOP), Var(BOOP)\n(l1.B < ?) & (l2.B > ?)'
        assert repr(c_n) == 'l1, l2 = NOT(BOOP), NOT(BOOP)\n(l1.B < ?) & (l2.B > ?)'

        c2 = NOT(l1.B < l2.B)
        assert repr(c2) == 'l1, l2 = NOT(BOOP), NOT(BOOP)\n(l1.B < l2.B)'
    # print(repr(c2))

def test_list_operations():
    with cre_context("test_list_operations"):
        TList = define_fact("TList",{"name" : "string", "items" : "ListType(string)"})
        v = Var(TList,"v")
        print(v.items[0])
        print(v.items[0] != v.items[1])
        assert ".items[0]" in str(v.items[0])

        c = v.items[0] != v.items[1]
        # assert str(c) == "~(v.items[0] == v.items[1])"
        assert str(c) == "AND(v:=Var(TList), ~(v.items[0] == v.items[1]))"

        Container = define_fact("Container",{"name" : "string", "children" : "ListType(Container)"})
        v = Var(Container,"v")
        assert ".children[0]" in str(v.children[0])

        c = v.children[0] != v.children[1]
        assert str(c) == "AND(v:=Var(Container), ~(v.children[0] == v.children[1]))"

@njit(cache=True)
def hsh(x):
    return hash(x)

def test_hash():
    ''' Tests hashing for Var, Literal, and Conditions '''
    with cre_context("test_hash"):
        TestLL = define_fact("TestLL",{"name": "string", "B" :'number', "nxt" : "TestLL"})

        ### VAR ### 
        a1 = Var(TestLL)
        a2 = a1
        b1 = Var(TestLL)

        assert hsh(a1) == hsh(a2)
        assert hsh(a1) != hsh(b1)

        a1 = Var(TestLL)
        b1 = Var(TestLL).nxt
        b2 = Var(TestLL).nxt.nxt.B

        assert hsh(a1) != hsh(b1)
        assert hsh(a1) != hsh(b2)

        ### LITERAL ### 
        x, y, z = Var(f8,'x'), Var(f8,'y'), Var(f8,'z')
        x2, y2, z2 = Var(f8,'x'), Var(f8,'y'), Var(f8,'z')

        a1 = literal_ctor((x + z) + (y + z))
        a2 = literal_ctor((x + z) + (y + z))
        b1 = literal_not(literal_ctor((x + z) + (y + z)))
        b2 = literal_ctor((x + z) + (y + x))
        b3 = literal_ctor((x2 + z2) + (y2 + z2))

        assert hsh(a1) == hsh(a2)
        assert hsh(a1) != hsh(b1)
        assert hsh(a1) != hsh(b2)
        assert hsh(a1) != hsh(b3)

        ### LITERAL w/ Consts ###

        a1 = literal_ctor((x + z) + (y + 1))
        a2 = literal_ctor((x + z) + (y + 1))
        b1 = literal_ctor((x + z) + (y + 7))

        assert hsh(a1) == hsh(a2)
        assert hsh(a1) != hsh(b1)

        ### CONDITIONS ### 
        X,Y = Var(TestLL,"X"), Var(TestLL,"Y")
        a1 = ( (X.B == 0) &
               ((X.nxt.B == Y.B) | (X.nxt.name == Y.name)) )
        a2 = ( (X.B == 0) &
               ((X.nxt.B == Y.B) | (X.nxt.name == Y.name)) )

        b1 = ( (A:=Var(TestLL)) & (A.B == 0) &
               (B:=Var(TestLL)) & ((A.nxt.B == B.B) | (A.nxt.name == B.name)) )
        b2 = ( (X.B < 0) &
               ((X.nxt.B == Y.B) | (X.nxt.name == Y.name)) )
        b3 = ( (X.B == 0) &
               ((X.nxt.B == Y.B)) )

        print(hsh(a1), hsh(a2), hsh(b1), hsh(b2), hsh(b3))

        assert hsh(a1) == hsh(a2)
        assert hsh(a1) != hsh(b1)
        assert hsh(a1) != hsh(b2)
        assert hsh(a1) != hsh(b3)


@njit(cache=True)
def eq(a,b):
    return _cast_structref(CREObjType, a)==_cast_structref(CREObjType, b)

def test_eq():
    with cre_context("test_eq"):
        TestLL = define_fact("TestLL",{"name": "string", "B" :'number', "nxt" : "TestLL"})

        ### VAR ### 
        a1 = Var(TestLL)
        a2 = a1
        b1 = Var(TestLL)

        assert eq(a1,a2)
        assert not eq(a1,b1)

        a1 = Var(TestLL)
        b1 = Var(TestLL).nxt
        b2 = Var(TestLL).nxt.nxt.B

        assert not eq(a1,b1)
        assert not eq(a1,b2)

        ### LITERAL ### 
        x, y, z = Var(f8,'x'), Var(f8,'y'), Var(f8,'z')
        x2, y2, z2 = Var(f8,'x'), Var(f8,'y'), Var(f8,'z')

        a1 = literal_ctor((x + z) + (y + z))
        a2 = literal_ctor((x + z) + (y + z))
        b1 = literal_not(literal_ctor((x + z) + (y + z)))
        b2 = literal_ctor((x + z) + (y + x))
        b3 = literal_ctor((x2 + z2) + (y2 + z2))

        assert eq(a1,a2)
        assert not eq(a1,b1)
        assert not eq(a1,b2)
        assert not eq(a1,b3)


        ### LITERAL w/ Consts ###

        a1 = literal_ctor((x + z) + (y + 1))
        a2 = literal_ctor((x + z) + (y + 1))
        b1 = literal_ctor((x + z) + (y + 7))

        assert eq(a1,a2)
        assert not eq(a1,b1)

        ### CONDITIONS ### 
        X,Y = Var(TestLL,"X"), Var(TestLL,"Y")
        a1 = ( (X.B == 0) &
               ((X.nxt.B == Y.B) | (X.nxt.name == Y.name)) )
        a2 = ( (X.B == 0) &
               ((X.nxt.B == Y.B) | (X.nxt.name == Y.name)) )

        b1 = ( (A:=Var(TestLL)) & (A.B == 0) &
               (B:=Var(TestLL)) & ((A.nxt.B == B.B) | (A.nxt.name == B.name)) )
        b2 = ( (X.B < 0) &
               ((X.nxt.B == Y.B) | (X.nxt.name == Y.name)) )
        b3 = ( (X.B == 0) &
               ((X.nxt.B == Y.B)) )

        assert eq(a1,a2)
        assert not eq(a1,b1)
        assert not eq(a1,b2)
        assert not eq(a1,b3)

from cre.conditions import conds_to_lit_sets, make_base_ptrs_to_inds, score_remaps
def test_anti_unify():
    x, y, z = Var(f8,'x'), Var(f8,'y'), Var(f8,'z')
    a, b, c, d = Var(f8,'a'), Var(f8,'b'), Var(f8,'c'), Var(f8,'d')

    # Single Conjunction Case
    c1 = (x < y) & (y < z) & (y < z) & (z != x) & (y != 0) 
    c2 = (a < b) & (b < c) & (b < c) & (b < c) & (c != a) & (b != 0) & (d != 0)
    c12_ref = ((x < y) & (y < z) & (y < z) & (z != x) & (y != 0))

    c12 = c1.antiunify(c2) 
    assert str(c12) == str(c12_ref)

    # Disjunction of Conjunctions Case
    c1 = ((x < y) & (z != x) & (y != 0) | # 1
          (x < y) & (z == x) & (y != 7) | # 2
          (x > y) & (z != x) & (y != 2)   # 3
         )
    c2 = ((a < b) & (c == a) & (b != 7) & (d > 0) | #2
          (a < b) & (c != a) & (b != 0) |           #1
          (a > b) & (c != a) & (b != 0) & (d != 7)  #3
         )

    c12_ref = ((x < y) & (z != x) & (y != 0) |\
              (x < y) & (z == x) & (y != 7) |\
              (x > y) & (z != x))

    c12, score = c1.antiunify(c2,return_score=True) #conds_antiunify(c1,c2)

    print(str(c12))
    print(str(c12_ref))
    assert str(c12) == str(c12_ref)
    assert score == 8./9.

    c1 = (x < y) & (y < z) & (y < z) & (z != x) & (y != 0) 
    c2 = (x < y) & (z < y) & (z < y) & (x != z) & (z != 0) 
    c12_ref = x & (x < y)

    c12, score = c1.antiunify(c2, return_score=True, fix_same_var=True) #conds_antiunify(c1,c2)

    print(str(c12))
    print(str(c12_ref))
    print_str_diff(str(c12),str(c12_ref))
    assert str(c12) == str(c12_ref)
    assert score == 1./5.

    X, Y, Z = Var(f8,'x'), Var(f8,'y'), Var(f8,'z')

    c1 = (x < y) & (y < z) & (y < z) & (z != x) & (y != 0) 
    c2 = (X < Y) & (Z < Y) & (Z < Y) & (X != Z) & (Z != 0) 
    c12_ref = x & (x < y)

    c12, score = c1.antiunify(c2, return_score=True, fix_same_alias=True) #conds_antiunify(c1,c2)
    assert str(c12) == str(c12_ref)
    assert score == 1./5.

    # TODO: Edge cases, unconditioned variables / variables with no supporting literals
    c1 = x & y & z
    c2 = X & Y & Z
    c12_ref = x & y & z
    c12, score = c1.antiunify(c2, return_score=True, fix_same_alias=True) #conds_antiunify(c1,c2)
    print("A", c12_ref)
    print("B", c12)
    print(score)
    # assert str(c12) == str(c12_ref)
    # assert score == 1.



    

if(__name__ == "__main__"):
    import faulthandler; faulthandler.enable()
    from cre.func import cre_func_unique_string
    x, y, z = Var(f8,'x'), Var(f8,'y'), Var(f8,'z')

    # print(cre_func_unique_string((x + z) + (y + z)))
    # print(cre_func_unique_string((x + z) + (y + 1)))
    # print(cre_func_unique_string((x + 1) + (y + z)))
    # test_var()
    test_anti_unify()
    # test_unconditioned()
    # test_build_conditions()
    # test_list_operations()
    # test_link()
    # test_initialize()
    # for i in range(10):
    #     t0 = time_ns()
    #     test_build_conditions()
    #     print(f'{(time_ns()-t0)/1e6} ms')
    # for i in range(10):
    #     t0 = time_ns()
    #     test_unconditioned()
    #     print(f'{(time_ns()-t0)/1e6} ms')
    # test_multiple_deref()
    # test_existential_not()
# bar.py_func()
    # bar()
    # test_hash()
    # test_eq()
    # 
    # exit()




import numpy as np
from numba import njit, f8, i8, u2, generated_jit
from numba.typed import List, Dict
from numba.types import DictType, ListType, unicode_type, Tuple
# from cre.op import Op
from cre.sc_planner import (gen_apply_multi_source, search_for_explanations,
                     apply_multi, SetChainingPlanner, insert_record,
                     join_records_of_type, forward_chain_one, extract_rec_entry,
                     retrace_goals_back_one, expl_tree_ctor, planner_declare,
                    build_explanation_tree, ExplanationTreeType, SC_Record, SC_RecordType)
from cre.utils import (_ptr_from_struct_incref, _list_from_ptr, _dict_from_ptr, _struct_from_ptr,
                        used_bytes, NRTStatsEnabled)
from cre.var import Var
from cre.context import cre_context
from cre.fact import define_fact
from cre.core import T_ID_FLOAT
from cre.func import CREFunc
# from cre.default_funcs import CastFloat, CastStr
from cre.default_funcs import CastFloat, CastStr
from numba.core.runtime.nrt import rtsys
import gc

import time
class PrintElapse():
    def __init__(self, name):
        self.name = name
    def __enter__(self):
        self.t0 = time.time_ns()/float(1e6)
    def __exit__(self,*args):
        self.t1 = time.time_ns()/float(1e6)
        print(f'{self.name}: {self.t1-self.t0:.2f} ms')


def get_base_funcs():
    from cre.default_funcs import Add, Multiply, Concatenate
    Add_f8 = Add(Var(f8),Var(f8))
    Multiply_f8 = Multiply(Var(f8),Var(f8))
    print(Add_f8, Multiply_f8)
    return Add_f8, Multiply_f8, Concatenate
    


i8_2x_tuple = Tuple((i8,i8))
def setup_float(planner=None,n=5):
    if(planner is None):
        planner = SetChainingPlanner()

    for x in range(5):
        planner.declare(float(x))

    return planner

def setup_str(planner=None,n=5):
    if(planner is None):
        planner = SetChainingPlanner()

    for x in range(65,n+65):
        planner.declare(chr(x))

    return planner

def test_apply_multi():
    Add, Multiply, Concatenate = get_base_funcs()
    planner = setup_float()
    rec = apply_multi(Add, planner, 0)
    d_typ = DictType(f8,i8_2x_tuple)
    f8_t_id = cre_context().get_t_id(f8)
    @njit(cache=True)
    def summary_vals_map(planner,target=6.0):
        d = _dict_from_ptr(d_typ, planner.val_map_ptr_dict[u2(f8_t_id)])
        return len(d), min(d), max(d)

    @njit(cache=True)
    def args_for(planner,target=6.0):
        d = _dict_from_ptr(d_typ, planner.val_map_ptr_dict[u2(f8_t_id)])
        l = List()
        re_ptr = d[target][1]
        # re_rec, re_next_re_ptr, re_args = extract_rec_entry(re_ptr)
        while(re_ptr != 0):
            re_rec, re_ptr, re_args = extract_rec_entry(re_ptr)
            l.append(re_args)
            # re_ptr = re_next_re_ptr
        return l
    assert summary_vals_map(planner) == (9,0.0,8.0)
    # print(np.array(args_for(planner,6)))
    assert np.array_equal(np.array(args_for(planner,6)),
                 np.array([[4, 2],[3, 3]]))
    

def test_insert_record():
    Add, Multiply, Concatenate = get_base_funcs()
    planner = setup_float()
    rec = apply_multi(Add, planner, 0)

    float64_t_id = cre_context().get_t_id(f8)
    insert_record(planner, rec, u2(float64_t_id), 1)
    @njit(cache=True)
    def len_f_recs(planner, t_id, depth):
        return len(planner.forward_records[depth][u2(t_id)])
    
    assert len_f_recs(planner,float64_t_id,1) == 1

@generated_jit(cache=True, nopython=True)
def summarize_depth_vals(planner, typ, depth):
    ''' Returns a summary of unique values of type 'typ' at 'depth':
        (len(flat_vals), min(flat_vals), max(flat_vals),
            len(val_map), min(val_map), max(val_map))
     '''
    from cre.fact import Fact
    _typ = typ.instance_type
    typ_name = str(_typ)
    typ_t_id = cre_context().get_t_id(_typ)
    l_typ = ListType(_typ)
    d_typ = DictType(_typ, Tuple((i8,i8)))
    if(isinstance(_typ, Fact)):
        def impl(planner, typ, depth): 
            print("----",typ_name, depth,"-----")
            tup = (u2(typ_t_id),depth)
            if(tup not in planner.flat_vals_ptr_dict): return 0,None, None, 0,None, None

            l = _list_from_ptr(l_typ, planner.flat_vals_ptr_dict[tup])
            d = _dict_from_ptr(d_typ, planner.val_map_ptr_dict[u2(typ_t_id)])
            print(l,d)
            first_l, last_l = None, None
            for i,x in enumerate(l):
                if(i == 0):
                    first_l = x
                if(i == len(l)-1):
                    last_l = x

            first_d, last_d = None, None
            for i,x in enumerate(d):
                if(i == 0):
                    first_d = x
                if(i == len(d)-1):
                    last_d = x


            return len(l), first_l, last_l, len(d), first_d, last_d
    else:
        def impl(planner, typ, depth): 

            print("----",typ_name, depth,"-----")
            tup = (u2(typ_t_id),depth)
            if(tup not in planner.flat_vals_ptr_dict): return 0,None, None, 0,None, None

            l = _list_from_ptr(l_typ, planner.flat_vals_ptr_dict[tup])
            d = _dict_from_ptr(d_typ, planner.val_map_ptr_dict[u2(typ_t_id)])
            # print(l,d)
            print(l)
            # print(d)
            return len(l), min(l),max(l),len(d), min(d),max(d)
    return impl

def test_join_records_of_type():
    Add, Multiply, Concatenate = get_base_funcs()
    planner = setup_float()
    rec = apply_multi(Add, planner, 0)
    insert_record(planner, rec, T_ID_FLOAT, 1)
    rec = apply_multi(Multiply, planner, 0)
    insert_record(planner, rec, T_ID_FLOAT, 1)

    # d_typ = DictType(f8, i8)
    # l_typ = ListType(f8)
    join_records_of_type(planner,1,f8)

    # @njit(cache=True)
    # def summarize_depth_vals(planner, typ_name, depth):
    #     l = _list_from_ptr(l_typ, planner.flat_vals_ptr_dict[(typ_name,depth)])
    #     d = _dict_from_ptr(d_typ, planner.val_map_ptr_dict[typ_name])
    #     return len(l), min(l),max(l),len(d), min(d),max(d)

    assert summarize_depth_vals(planner,f8, 1) == (12, 0.0, 16.0, 12, 0.0, 16.0)




def test_forward_chain_one():
    Add, Multiply, Concatenate = get_base_funcs()
    # fd_typ = DictType(f8, i8)
    # fl_typ = ListType(f8)
    # sd_typ = DictType(unicode_type, i8)
    # sl_typ = ListType(unicode_type)
    
    planner = setup_float()
    planner = setup_str(planner)
    forward_chain_one(planner, [Add,Multiply,Concatenate])

    assert summarize_depth_vals(planner,f8,1) == \
        (12, 0.0, 16.0, 12, 0.0, 16.0)

    assert summarize_depth_vals(planner,unicode_type,1) == \
        (30, 'A', 'EE', 30, 'A', 'EE')


    print("<<", planner.num_forward_inferences)

    forward_chain_one(planner, [Add,Multiply,Concatenate])


    print(summarize_depth_vals(planner,f8,1))
    assert summarize_depth_vals(planner,f8,2) == \
        (53, 0.0, 256.0, 53, 0.0, 256.0)

    print(summarize_depth_vals(planner,unicode_type,1))
    assert summarize_depth_vals(planner,unicode_type,2) == \
        (780, 'A', 'EEEE', 780, 'A', 'EEEE')


def setup_retrace(n=5):
    Add, Multiply, Concatenate = get_base_funcs()
    print(repr(Add), repr(Multiply), repr(Concatenate))
    planner = setup_float(n=n)
    planner = setup_str(planner,n=n)
    forward_chain_one(planner, [Add,Multiply,Concatenate])
    forward_chain_one(planner, [Add,Multiply,Concatenate])
    return planner


@njit(unicode_type(ExplanationTreeType,i8), cache=False)
def tree_str(root,ind=0):
    # print("START STR TREE")
    # if(len(root.children) == 0): return "?"
    s_ind = ' '*ind
    s = ''
    for entry in root.entries:
        # print("child.is_func", child.is_func)
        if(entry.is_func):
            func, child_arg_ptrs = entry.func, entry.child_arg_ptrs
        #     # for i in range(ind): s += " "
                
            s += f"\n{s_ind}{func}("
        #     # print(child_arg_ptrs)
            for i, ptr in enumerate(child_arg_ptrs):
                
                ch_expl = _struct_from_ptr(ExplanationTreeType, ptr)
                # print(ch_expl)
                # tree_str(ch_expl, ind+1)
        #         # print("str",tree_str(ch_expl, ind+1))
                s += f'{tree_str(ch_expl, ind+1)}'
                if (i != len(child_arg_ptrs) -1): s += ","
                    
                # s += ","
            s += ")"
        else:
            s += "?"
    return s        

def test_build_explanation_tree():
    planner = setup_retrace()
    print("BEF EX")
    root = build_explanation_tree(planner, f8, 36.0)
    print("BEF STR")
    for func in root:
        print(func)

def test_search_for_explanations(n=5):
    funcs = get_base_funcs()
    # print(repr(Add), repr(Multiply), repr(Concatenate))
    planner = setup_float(n=n)
    # planner = setup_str(planner,n=n)

    expl_tree = search_for_explanations(planner, 36.0, funcs=funcs, search_depth=2)
    # print(tree_str(expl_tree))
    for func in expl_tree:
        print(func)



#NOTE: Need to fix this seems to leak declared objects. 
def test_mem_leaks(n=5):
    with NRTStatsEnabled:
        with cre_context("test_mem_leaks") as context:
            funcs = get_base_funcs()
            init_used = used_bytes()

            # for i in range(5):
            #     planner = setup_float(n=n)
            #     expl_tree = search_for_explanations(planner, 36.0,
            #         funcs=funcs, search_depth=2, context=context)
            #     expl_tree_iter = iter(expl_tree)
            #     for f_comp,binding in expl_tree_iter:
            #         pass

            #     planner = None
            #     expl_tree = None
            #     expl_tree_iter = None
            #     if(i == 0): 
            #         init_used = used_bytes()
            #     else:
            #         print(used_bytes() - init_used)

            # assert used_bytes() == init_used


            BOOP = define_fact("BOOP", {
                "A" : "string",
                "B" : {"type": "number", "visible":  True, "semantic" : True}
            })
            
            def declare_em(planner,s="A"):
                for i in range(n):
                    b = BOOP(s,i)
                    print("BEF", b._meminfo.refcount)
                    planner.declare(b)
                    print("AFT", b._meminfo.refcount)
                return b

            print("-----")
            for i in range(5):
                planner = SetChainingPlanner([BOOP])
                b = declare_em(planner,"A")
                # expl_tree = search_for_explanations(planner, 36.0,
                #     funcs=funcs, search_depth=2, context=context)
                # expl_tree_iter = iter(expl_tree)
                # for f_comp,binding in expl_tree_iter:
                #     pass
                print("<<", b._meminfo.refcount)

                planner = None
                expl_tree = None
                expl_tree_iter = None
                b = None
                if(i == 0): 
                    init_used = used_bytes()
                else:
                    print(used_bytes() - init_used)

            assert used_bytes() == init_used




def test_declare_fact():
    with cre_context("test_declare_fact"):
        # print("vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv")
        BOOP = define_fact("BOOP", {
            "A" : "string",
            "B" : {"type": "number", "visible":  True, "semantic" : True}
        })
        
        def declare_em(planner,s="A"):
            for i in range(5):
                b = BOOP(s,i)
                planner.declare(b)

        planner = SetChainingPlanner([BOOP])
        declare_em(planner,"A")

        # print("vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv")
        # print(summarize_depth_vals(planner, BOOP, 0))
        # print(summarize_depth_vals(planner, unicode_type, 0))
        # print(summarize_depth_vals(planner, f8, 0))
        # print("^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^")
        assert summarize_depth_vals(planner, BOOP, 0)[0] == 5
        # assert summarize_depth_vals(planner, unicode_type, 0)[0] == 5
        assert summarize_depth_vals(planner, f8, 0)[0] == 5


        expls = planner.search_for_explanations(36.0, funcs=get_base_funcs(), search_depth=2)
        A_f_comp_binding_pairs = list(iter(expls))

        planner = SetChainingPlanner([BOOP])
        declare_em(planner,"A")
        declare_em(planner,"B")
        
        assert summarize_depth_vals(planner, BOOP, 0)[0] == 10
        # assert summarize_depth_vals(planner, unicode_type, 0)[0] == 5
        assert summarize_depth_vals(planner, f8, 0)[0] == 5

        expls = planner.search_for_explanations(36.0, funcs=get_base_funcs(), search_depth=2)
        AB_f_comp_binding_pairs = list(iter(expls))

        assert len(AB_f_comp_binding_pairs) >= 4 * len(A_f_comp_binding_pairs)

        # print("(()()))")
        for i, (func, binding) in enumerate(A_f_comp_binding_pairs):
            print("<<", func, binding)


            assert(func(*binding)==36.0)

            # Flattening takes a while so stop after 3 
            if(i >= 2): break
            

def test_declare_fact_w_conversions():
    with cre_context("test_declare_fact_w_conversions"):
        BOOP = define_fact("BOOP", {
            "A" : str,
            "B" : {"type": str, "visible":  True,
                 "semantic" : True, 'conversions' : {float : CastFloat}}
        })
        
        def declare_em(planner,s="A"):
            for i in range(5):
                b = BOOP(s,str(i))
                planner.declare(b)

            # Declare one that can't be converted
            planner.declare(BOOP(s,s))

        planner = SetChainingPlanner([BOOP])
        declare_em(planner,"A")
        print("---------------------------------------------------------")

        assert summarize_depth_vals(planner, BOOP, 0)[0] == 6
        assert summarize_depth_vals(planner, unicode_type, 0)[0] == 6
        assert summarize_depth_vals(planner, f8, 0)[0] == 5

        expls = planner.search_for_explanations(36.0, funcs=get_base_funcs(), search_depth=2)
        A_f_comp_binding_pairs = list(iter(expls))

        assert len(A_f_comp_binding_pairs) > 0

        print("---------------------------------------------------------")

        planner = SetChainingPlanner([BOOP])
        declare_em(planner,"A")
        declare_em(planner,"B")
        
        assert summarize_depth_vals(planner, BOOP, 0)[0] == 12
        assert summarize_depth_vals(planner, unicode_type, 0)[0] == 7
        assert summarize_depth_vals(planner, f8, 0)[0] == 5

        expls = planner.search_for_explanations(36.0, funcs=get_base_funcs(), search_depth=2)
        AB_f_comp_binding_pairs = list(iter(expls))

        assert len(AB_f_comp_binding_pairs) >= 4 * len(A_f_comp_binding_pairs)

        for i, (func, binding) in enumerate(A_f_comp_binding_pairs):
            print("<<", func, binding)

            assert(func(*binding)==36.0)

            # Flattening takes a while so stop after 3 
            if(i >= 2): break

        print("---------------------------------------------------------")
        # Check for key error bug when don't have funcs for all decalared types.
        from cre.default_funcs import Add, Multiply
        Add_f8 = Add(f8, f8)
        Multiply_f8 = Multiply(f8, f8)
        funcs = [Add_f8, Multiply_f8]
        planner = SetChainingPlanner([BOOP])
        
        for i in range(5):
            b = BOOP("A",str(i))
            planner.declare(b)

        expls = planner.search_for_explanations(36.0, funcs=funcs, search_depth=2)
        AB_f_comp_binding_pairs = list(iter(expls))

        assert len(AB_f_comp_binding_pairs) > 0

        for i, (func, binding) in enumerate(AB_f_comp_binding_pairs):
            print("<<", func, binding)

            assert(func(*binding)==36.0)

            # Flattening takes a while so stop after 3 
            if(i >= 2): break


        # Special Example
        # TODO: WRITE ASSERTIONS FOR THIS
        @CREFunc(signature=f8(f8,f8),
            shorthand = '{0} + {1}',
            commutes=True)
        def Add(a, b):
            return a + b

        @CREFunc(signature=f8(f8,f8,f8),
            shorthand = '{0} + {1} + {2}',
            commutes=True)
        def Add3(a, b, c):
            return a + b + c

        @CREFunc(signature=f8(f8), shorthand = 'TensDigit({0})')
        def TensDigit(a):
            return (a // 10) % 10

        print("THIS", TensDigit(12.0), TensDigit(7))

        @CREFunc(signature=f8(f8), shorthand = 'OnesDigit({0})')
        def OnesDigit(a):
            return a % 10

        funcs = [OnesDigit, TensDigit, Add, Add3]
        planner = SetChainingPlanner([BOOP])
        planner.declare(BOOP("A","7"))
        planner.declare(BOOP("B","7"))
        planner.declare(BOOP("C","1"))
        expls = planner.search_for_explanations(1.0, funcs=funcs, search_depth=2, min_stop_depth=2, min_solution_depth=1)
        if(expls is None):
            print("NO EXPLANATIONS")
        else:
            for f in expls:
                print(f)
        print(summarize_depth_vals(planner, f8, 0))
        print(summarize_depth_vals(planner, f8, 1))
        print(summarize_depth_vals(planner, f8, 2))


        funcs = [OnesDigit, TensDigit, Add3]
        planner = SetChainingPlanner([BOOP])
        planner.declare(BOOP("A","7"))
        planner.declare(BOOP("B","7"))
        planner.declare(BOOP("C","1"))
        expls = planner.search_for_explanations(5.0, funcs=funcs, search_depth=2, min_stop_depth=1)
        if(expls is None):
            print("NO EXPLANATIONS")
        else:
            for f in expls:
                print(f)
        print(summarize_depth_vals(planner, f8, 0))
        print(summarize_depth_vals(planner, f8, 1))
        print(summarize_depth_vals(planner, f8, 2))


        # AB_f_comp_binding_pairs = list(iter(expls))




def test_min_stop_depth():
    with cre_context("test_min_stop_depth"):
        BOOP = define_fact("BOOP", {
            "A" : str,
            "B" : {"type": str, "visible":  True,
                 "semantic" : True, 'conversions' : {float : CastFloat}}
        })
        from cre.default_funcs import Add, Multiply
        Add_f8 = Add(f8, f8)
        Multiply_f8 = Multiply(f8, f8)

        planner = SetChainingPlanner([BOOP])

        for i in range(5):
            b = BOOP("A",str(i))
            planner.declare(b)

        expls = planner.search_for_explanations(36.0, funcs=[Add_f8, Multiply_f8], search_depth=2)
        new_func, match = list(expls)[0]
        print(new_func, match)
        # new_func = f_comp.flatten()

        planner = SetChainingPlanner([BOOP])

        for i in range(5):
            b = BOOP("A",str(i))
            planner.declare(b)
        planner.declare(BOOP("Q","36"))
        print("----------------------------------------")

        expls = planner.search_for_explanations(36.0, funcs=[new_func], 
            search_depth=1, min_stop_depth=1)
        assert len(list(expls)) > 0

        func, match = list(expls)[0]

        print(func, match)

        planner = SetChainingPlanner([BOOP])
        for i in range(2,5):
            b = BOOP("A",str(i))
            planner.declare(b)

        expls = planner.search_for_explanations(3.0, funcs=[new_func], 
            search_depth=1, min_stop_depth=1, min_solution_depth=1)

        assert expls == None

        print("----------------------------------------")
        planner = SetChainingPlanner([BOOP])
        for i in range(5):
            b = BOOP("A",str(i))
            planner.declare(b)
        # planner.declare(BOOP("Q","36"))

        expls = planner.search_for_explanations(36.0, funcs=[Add_f8, Multiply_f8],
                    min_stop_depth=1, search_depth=1)

        print(summarize_depth_vals(planner, BOOP, 1))
        print(summarize_depth_vals(planner, unicode_type, 1))
        print(summarize_depth_vals(planner, f8, 1))
        # for i, (f_comp, binding) in enumerate(expls):
        #     print(f_comp, binding)
        # raise ValueError()

        print("END")

        expls = planner.search_for_explanations(36.0, funcs=[Add_f8, Multiply_f8],
                    min_stop_depth=1, search_depth=2)

        print(summarize_depth_vals(planner, BOOP, 2))
        print(summarize_depth_vals(planner, unicode_type, 2))
        print(summarize_depth_vals(planner, f8, 1))
        print(summarize_depth_vals(planner, f8, 2))

        all_expls =  list(iter(expls))
        print(len(all_expls))

        # raise ValueError()

        for i, (func, match) in enumerate(expls):
            print(func, match)
            print(func(*match))

        print("END")



def test_non_numerical_vals():
    with cre_context("test_min_stop_depth"):
        BOOP = define_fact("BOOP", {
            "A" : str,
            "B" : {"type": str, "visible":  True,
                 "semantic" : True, 'conversions' : {float : CastFloat}}
        })
        from cre.default_funcs import Add, Multiply
        Add_f8 = Add(f8, f8)

        planner = SetChainingPlanner([BOOP])
        planner.declare(BOOP("A",'1'))
        planner.declare(BOOP("B",'+'))
        planner.declare(BOOP("C",'1'))

        print(summarize_depth_vals(planner,BOOP, 0))
        print(summarize_depth_vals(planner,unicode_type, 0))
        print(summarize_depth_vals(planner,f8, 0))

        expls = planner.search_for_explanations(2.0, funcs=[Add_f8], 
            search_depth=1)

        for i, (func, binding) in enumerate(expls):
            print(func, binding)

def test_const_funcs():
    with cre_context("test_const_funcs"):
        BOOP = define_fact("BOOP", {
            "A" : str,
            "B" : {"type": str, "visible":  True,
                 "semantic" : True, 'conversions' : {float : CastFloat}}
        })
        from cre.default_funcs import Add, Multiply
        Add_f8 = Add(f8, f8)

        planner = SetChainingPlanner([BOOP])
        planner.declare(BOOP("A",'1'))
        planner.declare(BOOP("B",'+'))
        planner.declare(BOOP("C",'1'))

        @CREFunc(signature=f8(), no_raise=True, shorthand='10')
        def Ten():
            return 10

        print(summarize_depth_vals(planner,BOOP, 0))
        print(summarize_depth_vals(planner,unicode_type, 0))
        print(summarize_depth_vals(planner,f8, 0))

        expls = planner.search_for_explanations(11.0, funcs=[Add_f8, Ten], 
            search_depth=1)

        for i, (func, match) in enumerate(expls):
            print(func, match)
            print(func(*match))
            assert func(*match) == 11.0
            


def test_policy_search(n=5):
    [Add_f8, Multiply_f8, Concatenate] = funcs = get_base_funcs()
        
    # No Policy
    planner = setup_float(n=n)
    expl_tree = search_for_explanations(planner, 36.0, funcs=funcs, search_depth=2)
    no_policy_expls = list(expl_tree)    

    # Policy
    policy = [[Add_f8],[Multiply_f8]]
    planner = setup_float(n=n)
    expl_tree = search_for_explanations(planner, 36.0, policy=policy, search_depth=2)
    policy_expls = list(expl_tree)    

    for expl in no_policy_expls:
        print(expl)
    print("----------------", len(no_policy_expls))
    for i, expl in enumerate(set([str(x) for x in no_policy_expls])):
        print(i, expl)        
    print("----------------")
    for expl in policy_expls:
        print(expl)

    # Policy w/ Args
    policy = [[(Add_f8, [4.0,2.0])],[(Multiply_f8, [])]]
    planner = setup_float(n=n)
    expl_tree = search_for_explanations(planner, 36.0, policy=policy, search_depth=2)
    policy_expls = list(expl_tree)    
    
    assert len(policy_expls) < len(no_policy_expls)

def test_divide():
    from cre.default_funcs import Divide, Multiply
    Divide_f8 = Divide(Var(f8),Var(f8))
    Multiply_f8 = Multiply(Var(f8),Var(f8))


    planner = SetChainingPlanner()
    for x in [360, 135, 6]:
        planner.declare(float(x))

    expls = planner.search_for_explanations(13.5, funcs=[Divide_f8, Multiply_f8], 
            search_depth=3, min_stop_depth=1)
    for f, m in expls:
        print(f.depth, f, m)


    planner = SetChainingPlanner()
    for x in [2, 12, 3]:
        planner.declare(float(x))

    expls = planner.search_for_explanations(8.0, funcs=[Divide_f8, Multiply_f8], 
            search_depth=3, min_stop_depth=1)

    print(summarize_depth_vals(planner, f8, 0))
    print(summarize_depth_vals(planner, f8, 1))
    print(summarize_depth_vals(planner, f8, 2))

    for f, m in expls:
        print(f.depth, f, m)


def test_const_declarations():
    from cre.default_funcs import Divide, Multiply
    Divide_f8 = Divide(Var(f8),Var(f8))
    Multiply_f8 = Multiply(Var(f8),Var(f8))


    planner = SetChainingPlanner()
    planner.declare(float(360), is_const=True)
    for x in [135, 6]:
        planner.declare(float(x))

    # Test No Policy
    expls = planner.search_for_explanations(13.5, funcs=[Divide_f8, Multiply_f8], 
            search_depth=3, min_stop_depth=1)
    for f, m in expls:
        print(f.depth, f, m)


    planner = SetChainingPlanner()
    planner.declare(float(360), is_const=True)
    for x in [135, 6]:
        planner.declare(float(x))

    # Test Policy
    policy = [[(Divide_f8, [135.0, 360.0]), (Multiply_f8, [6.0, 6.0])], [(Multiply_f8, [])]]
    expls = planner.search_for_explanations(13.5, policy=policy, 
            search_depth=3, min_stop_depth=1)
    for f, m in expls:
        print(f.depth, f, m)

    # Another test
    planner = SetChainingPlanner()
    for x in [2, 12, 3]:
        planner.declare(float(x))

    expls = planner.search_for_explanations(8.0, funcs=[Divide_f8, Multiply_f8], 
            search_depth=3, min_stop_depth=1)

    print(summarize_depth_vals(planner, f8, 0))
    print(summarize_depth_vals(planner, f8, 1))
    print(summarize_depth_vals(planner, f8, 2))

    for f, m in expls:
        print(f.depth, f, m)





        # planner = SetChainingPlanner([BOOP])
        # planner.declare(7.0)
        # planner.declare(1.0)
        # planner.declare(6.0)
        # planner.declare(9.0)
        # planner.declare(5.0)
        # planner.declare(4.0)
        # expls = planner.search_for_explanations(36.0, funcs=funcs, search_depth=2)







        # gen_src_declare_fact(BOOP, ["A","B"])
        # gen_src_declare_fact(BOOP, ["A","B"])

        # planner = SetChainingPlanner()
        # b = BOOP("A",1.0)
        # planner_declare_fact(planner,b,[("B","unicode_type"), ("B", 'float')])


    # goals = Dict.empty(f8,ExplanationTreeType)
    # goals[36.0] = expl_tree_ctor()
    # retrace_goals_back_one(planner, goals)
    # goals = List([36.0])
    # e_trees = List([expl_tree_ctor()])
    # print(retrace_back_one(planner, DictType(f8,i8),'float64', goals, e_trees))

    # goals = List(["AABC"])
    # e_trees = List([expl_tree_ctor()])
    # print(retrace_back_one(planner, DictType(unicode_type,i8),'unicode_type', goals, e_trees))

    # build_explanation_tree(planner,36.0, f8)


def benchmark_apply_multi():
    Add, Multiply, Concatenate = get_base_funcs()
    planner = setup_float(n=1000)

    apply_multi(Add, planner, 0)
    with PrintElapse("benchmark_apply_multi"):
        for i in range(10):
            apply_multi(Add, planner, 0)

def benchmark_retrace_goals_back_one():
    Add, Multiply, Concatenate = get_base_funcs()
    planner = setup_retrace()
    goals = List([36.0])

    apply_multi(Add, planner, 0)
    with PrintElapse("benchmark_retrace_back_one"):
        for i in range(10):
            retrace_goals_back_one(planner, DictType(f8,i8),'float64', goals)



# @njit(cache=False)
# def foo_gen():
#     for i in range(10):
#         yield i

# def product_of_generators(generators):
#     iters = []
#     out = []
    
#     while(True):
#         #Create any iterators that need to be created
#         while(len(iters) < len(generators)):
#             it = generators[len(iters)]()
#             iters.append(it)
        
#         iter_did_end = False
#         while(len(out) < len(iters)):
#             #Try to fill in any missing part of out
#             try:
#                 nxt = next(iters[len(out)])
#                 out.append(nxt)
#             #If any of the iterators failed pop up an iterator
#             except StopIteration as e:
#                 # Stop yielding when 0th iter fails
#                 if(len(iters) == 1):
#                     return
#                 out = out[:-1]
#                 iters = iters[:-1]
#                 iter_did_end = True

#         if(iter_did_end): continue

#         yield out
#         out = out[:-1]

# with PrintElapse("gen_iters"):
#     l = [x for x in product_of_generators([foo_gen,foo_gen,foo_gen, foo_gen])]
#     print(len(l))
#     print()





if __name__ == "__main__":
    # Makes it easier to track down segfaults
    import faulthandler; faulthandler.enable()
    # test_non_numerical_vals()
    # with PrintElapse("test_build_explanation_tree"):
    #     test_build_explanation_tree()
    # with PrintElapse("test_build_explanation_tree"):
    #     test_build_explanation_tree()
    # with PrintElapse("test_search_for_explanations"):
    #     test_search_for_explanations()
    # with PrintElapse("test_search_for_explanations"):
    #     test_search_for_explanations()
# 

    # pass
    # test_apply_multi()
    # test_insert_record()
    # test_join_records_of_type()
    # test_forward_chain_one()
    # test_build_explanation_tree()
    # test_search_for_explanations()
    # test_divide()
    # test_declare_fact()
    # test_mem_leaks(n=10)
    # benchmark_apply_multi()
    # benchmark_retrace_back_one()
        # test_apply_multi()
    # gen = foo_gen()
    # for i in gen:
    #     print(i)
    # test_declare_fact()
    # test_declare_fact()
    # test_declare_fact_w_conversions()
    test_min_stop_depth()
    # test_const_funcs()
    # test_const_declarations()

    # test_policy_search()
# from numba import njit, i8
# from numba.typed import Dict
# from numba.types import ListType
# import numpy as np
# import dill
# from cre.utils import _struct_from_ptr, _ptr_from_struct_incref
# from cre.condensed_chainer import CondensedRecord
# from cre_cache.Add._5e24697b8e500d3d837dca80591bde623483d2322c5204e56fd36c79ddc2ed7d import call, check

# typ0, = dill.loads(b'\x80\x04\x95\xd5\x00\x00\x00\x00\x00\x00\x00\x8c\x19numba.core.types.abstract\x94\x8c\x13_type_reconstructor\x94\x93\x94\x8c\x07copyreg\x94\x8c\x0e_reconstructor\x94\x93\x94\x8c\x18numba.core.types.scalars\x94\x8c\x05Float\x94\x93\x94\x8c\ndill._dill\x94\x8c\n_load_type\x94\x93\x94\x8c\x06object\x94\x85\x94R\x94N\x87\x94}\x94(\x8c\x04name\x94\x8c\x07float64\x94\x8c\x08bitwidth\x94K@\x8c\x05_code\x94K\x17u\x87\x94R\x94\x85\x94.')
# ret_typ = dill.loads(b'\x80\x04\x95\xd3\x00\x00\x00\x00\x00\x00\x00\x8c\x19numba.core.types.abstract\x94\x8c\x13_type_reconstructor\x94\x93\x94\x8c\x07copyreg\x94\x8c\x0e_reconstructor\x94\x93\x94\x8c\x18numba.core.types.scalars\x94\x8c\x05Float\x94\x93\x94\x8c\ndill._dill\x94\x8c\n_load_type\x94\x93\x94\x8c\x06object\x94\x85\x94R\x94N\x87\x94}\x94(\x8c\x04name\x94\x8c\x07float64\x94\x8c\x08bitwidth\x94K@\x8c\x05_code\x94K\x17u\x87\x94R\x94.')

# l_typ0 = ListType(typ0)
# @njit(cache=True)
# def Add_apply_multi(planner, depth, start0=0, start1=0):
#     tup0 = ('float64',depth)
#     print("START")
#     if(tup0 in planner.flat_vals_ptr_dict):
#         iter_ptr0 = planner.flat_vals_ptr_dict[tup0]
#         iter0 = _list_from_ptr(l_typ0, iter_ptr0)
#         print(iter0)
#     else:
#         return None
#     print("End")
#     for x in iter0:
#         print(x)
#     print("?",len(iter0))
#     l0, l1 = len(iter0)-start0, len(iter0)-start1
#     hist_shape = (l0, l1)
#     print(hist_shape )
#     hist = np.zeros(hist_shape, dtype=np.uint64)
#     vals_to_uid = Dict.empty(ret_typ, i8)
#     print(hist)
#     uid=1
#     for i0 in range(start0,len(iter0)):
#         for i1 in range(start1,len(iter0)):
#             print(i0, i1)
#             a0,a1 = iter0[i0],iter0[i1]
#             if(not check(a0,a1)): continue
#             v = call(a0,a1)
#             if(v in vals_to_uid):
#                 hist[i0,i1] = vals_to_uid[v]
#             else:
#                 hist[i0,i1] = uid
#                 vals_to_uid[v] = uid; uid+=1
#     vals_to_uid_ptr = _ptr_from_struct_incref(vals_to_uid)
#     return CondensedRecord(hist.flatten(), hist_shape, vals_to_uid_ptr)
    

# print(apply_multi(Add,planner, 0))
# print(





import operator
import numpy as np
import numba
from numba import types, njit, i8, u8, i4, u1,u2,u4,  i8, literally, generated_jit, objmode
from numba.typed import List
from numba.core.types import ListType, unicode_type, void, Tuple
from numba.experimental import structref
from numba.experimental.structref import new, define_attributes, _Utils
from numba.extending import SentryLiteralArgs, lower_cast, overload_method, intrinsic, overload_attribute, intrinsic, lower_getattr_generic, overload, infer_getattr, lower_setattr_generic
from numba.core.typing.templates import AttributeTemplate
from cre.context import cre_context, get_cre_context_data
from cre.structref import define_structref, define_boxing, define_structref_template, CastFriendlyStructref
from cre.fact import define_fact, BaseFact, cast_fact, DeferredFactRefType, Fact, _standardize_type
from cre.utils import cast, PrintElapse, ptr_t, decode_idrec, lower_getattr,  lower_setattr, lower_getattr, _decref_ptr, _incref_ptr, _incref_structref, _ptr_from_struct_incref
from cre.utils import assign_to_alias_in_parent_frame, encode_idrec, _obj_cast_codegen
from cre.vector import VectorType
from cre.obj import cre_obj_field_dict,CREObjType, CREObjTypeClass, CREObjProxy, set_chr_mbrs
from cre.type_conv import ptr_to_var_name
# from cre.predicate_node import BasePredicateNode,BasePredicateNodeType, get_alpha_predicate_node_definition, \
# get_beta_predicate_node_definition, deref_attrs, define_alpha_predicate_node, define_beta_predicate_node, AlphaPredicateNode, BetaPredicateNode
from numba.core import imputils, cgutils
from numba.core.datamodel import default_manager, models
from numba.experimental.structref import StructRefProxy


from operator import itemgetter
from copy import copy
from os import getenv
from cre.utils import deref_info_type, DEREF_TYPE_ATTR, DEREF_TYPE_LIST, listtype_sizeof_item, _obj_cast_codegen
from cre.core import T_ID_VAR, register_global_default
# import inspect



token_fields_dict = {
    "t_id" : u4,
    "length" : u4,
    "value" : unicode_type,
}

class TokenTypeClass(CREObjTypeClass):
    t_id = T_ID_VAR
    type_cache = {}

    

# @lower_cast(VarTypeClass, CREObjType)
# def upcast(context, builder, fromty, toty, val):
#     return _obj_cast_codegen(context, builder, val, fromty, toty,incref=False)



# Manually register the type to avoid automatic getattr overloading 
default_manager.register(VarTypeClass, models.StructRefModel)

VarType = VarTypeClass()
register_global_default("Var", VarType)

# Allow typed Var instances to be upcast to VarType
@lower_cast(VarTypeClass, VarType)
def upcast(context, builder, fromty, toty, val):
    return _obj_cast_codegen(context, builder, val, fromty, toty,incref=False)

class Var(StructRefProxy):
    t_id = T_ID_VAR
    def __new__(cls, typ, alias="", skip_assign_alias=False):
        # if(not isinstance(typ, types.StructRef)): typ = typ.fact_type
        typ = _standardize_type(typ, cre_context())
        # base_type_name = str(typ)
        # print(base_type_name)
                
        base_t_id = cre_context().get_t_id(_type=typ)

        if(getenv("CRE_SPECIALIZE_VAR_TYPE",default=False)):
            raise ValueError("THIS SHOULDN'T HAPPEN")
            typ_ref = types.TypeRef(typ)
            struct_type = VarTypeClass(typ_ref,typ_ref)
            st = var_ctor(struct_type, base_t_id, alias)
        else:
            st = var_ctor_generic(base_t_id, alias)
        
        st._base_type = typ
        st._head_type = typ
        st._derefs_str = ""

        # if(alias):
            # import inspect, ctypes
            # if(alias is not None): 
            #     # Binds this instance globally in the calling python context 
            #     #  so that it is bound to a variable named whatever alias was set to
            #     # print(inspect.stack()[2][0].f_locals)
            #     # Get the calling frame
            #     frame = inspect.stack()[2][0] 
            #     # Assign the Var to it's alias
            #     frame.f_locals[alias] = st
            #     # Update locals()
            #     ctypes.pythonapi.PyFrame_LocalsToFast(ctypes.py_object(frame), ctypes.c_int(1))
            # # assign_to_alias_in_parent_frame(st, alias)

        return st
        
   

# Manually define the boxing to avoid constructor overloading
define_boxing(VarTypeClass, Var)



from numba import types

attr_filter_registry = {}

class AttrFilter():
    def __init__(self, flag, filter_func=None):
        self.flag = flag
        if(filter_func is None): 
            filter_func = lambda attr, config, _ : False
        self.filter_func = filter_func
        attr_filter_registry[self.flag] = self

    def get_attrs(self, fact_type, negated=False):
        clean_spec = fact_type.clean_spec

        for attr, config in clean_spec.items():
            flag = config.get(self.flag, None)
            if(flag == True ^ negated): yield attr, config
            if(flag == False ^ negated): continue
            if(self.filter_func(attr, config, clean_spec) ^ negated):
                yield attr, config

def should_be_relational(attr, config, clean_spec):
    from cre.fact import Fact
    attr_t = config['type']
    if(isinstance(attr_t, Fact)): return True
    if(isinstance(attr_t, types.ListType)):
        if(isinstance(attr_t.item_type, Fact)): return True
    return False

def should_be_spatial(attr, config, clean_spec):
    return (
        (attr_t == "x" and isinstance(config['type'], types.number_domain)) or 
        (attr_t == "y" and isinstance(config['type'], types.number_domain)) or 
        (attr_t == "width" and isinstance(config['type'], types.number_domain)) or 
        (attr_t == "height" and isinstance(config['type'], types.number_domain))
    )

def should_be_few_valued(attr, config, clean_spec):
    return isinstance(config['type'], types.Boolean)

def should_be_parent(attr, config, clean_spec):
    return "parent" in attr 

def should_be_unique_id(attr, config, clean_spec):
    if("id" in clean_spec):
        return attr == "id"
    elif("name" in clean_spec):
        return attr == "name"
    return False

visible = AttrFilter("visible")
relational = AttrFilter("relational", should_be_relational)
semantic = AttrFilter("semantic")
spatial = AttrFilter("spatial", should_be_spatial)
few_valued = AttrFilter("few_valued", should_be_few_valued)
parent = AttrFilter("parent", should_be_few_valued)
unique_id = AttrFilter("unique_id", should_be_unique_id)



import contextvars


skip = contextvars.ContextVar("cre_skip_package_level_imports",default=False)
# print("<<", skip.get())
if(not skip.get()):
    import cre.cfuncs

    # Instantiate the default context at the package level
    from cre.version import __version__
    import cre.context as context_module

    cre_context_ctxvar = contextvars.ContextVar("cre_context",
            default=context_module.CREContext.get_default_context()) 
    context_module.cre_context_ctxvar = cre_context_ctxvar

    # Necessary to import these at package level to ensure that they are defined
    #  before a context is created
    import cre.obj
    from cre.fact import define_fact, BaseFact, Fact, FactProxy
    from cre.tuple_fact import TupleFact, TF
    from cre.var import Var
    from cre.func import CREFunc, UntypedCREFunc
    from cre.conditions import Literal, Conditions

    # Other helpful things exposed at the package level
    from cre.context import cre_context
    from cre.memset import MemSet

    import cre.dynamic_exec

    # Import this to monkey-patch numba 
    import cre.type_conv

    



import numpy as np
from numba import njit, u8, types
from numba.cpython.hashing import (_Py_hash_t, _Py_uhash_t, _PyHASH_XXROTATE,
        _PyHASH_XXPRIME_1, _PyHASH_XXPRIME_2, _PyHASH_XXPRIME_5, process_return,
        _siphash24)
from numba.core.extending import register_jitable

# ------------------------------------------------------------
# : accum_item_hash

@njit(_Py_uhash_t(_Py_uhash_t,_Py_uhash_t), cache=True)
def accum_item_hash(acc, lane):
    if lane == _Py_uhash_t(-1):
        return _Py_uhash_t(1546275796)
    acc += lane * _PyHASH_XXPRIME_2
    acc = _PyHASH_XXROTATE(acc)
    acc *= _PyHASH_XXPRIME_1
    return acc


# ------------------------------------------------------------
# : unicode_hash_noseed

from numba.core.unsafe.bytes import grab_byte, grab_uint64_t
from numba.cpython.unicode import _kind_to_byte_width
import sys

# Set a biggish hash-cutoff for using fast DJBX33A, instead of slow siphash 
#  should speed things up appreciably, security sensitive things shouldn't
#  be calling the no_seed hash implementation anyway.
_HASH_CUTOFF = 10

# Constants instead of using Python's random seed
const_siphash_k0 = u8(7823388449531244853)
const_siphash_k1 = u8(2452911659039219235)

@register_jitable(locals={'_hash': _Py_uhash_t})
def _Py_HashBytes_noseed(val, _len):
    if (_len == 0):
        return process_return(0)

    if (_len < _HASH_CUTOFF):
        # TODO: this branch needs testing, needs a CPython setup for it!
        # /* Optimize hashing of very small strings with inline DJBX33A. */
        _hash = _Py_uhash_t(5381)  # /* DJBX33A starts with 5381 */
        for idx in range(_len):
            _hash = ((_hash << 5) + _hash) + np.uint8(grab_byte(val, idx))

        _hash ^= _len
    else:
        tmp = _siphash24(types.uint64(const_siphash_k0),
                         types.uint64(const_siphash_k1),
                         val, _len)
        _hash = process_return(tmp)
    return process_return(_hash)

# Hashes a unicode string, but without using python's hash seed.
#  Ensures that the hash is is the same between executions. 
@njit(cache=True)
def unicode_hash_noseed(val):    
    kindwidth = _kind_to_byte_width(val._kind)
    _len = len(val)
    return _Py_HashBytes_noseed(val._data, kindwidth * _len)



import numpy as np
from cre.fact import (_fact_from_spec, _standardize_spec, _merge_spec_inheritance, 
     define_fact, cast_fact, BaseFact, DeferredFactRefType, isa,
      uint_to_inheritance_bytes, get_inheritance_bytes_len_ptr, get_inheritance_t_ids)
from cre.utils import cast
from cre.context import cre_context
from cre.memset import MemSet
from cre.obj import CREObjType, copy_cre_obj
from numba import njit, u8, u1, generated_jit
from numba.typed import List
from numba.types import ListType
import cre.dynamic_exec
import pytest
import operator
import cloudpickle
import pytest

def test__standardize_spec():
    with cre_context("test__standardize_spec") as context:
        spec = {"A" : "string", "B" : "number"}
        spec = _standardize_spec(spec, context,"BOOP")
        print(spec)

        #Standardized specs should at least have 'type'
        assert str(spec['A']['type']) == 'unicode_type'
        assert str(spec['B']['type']) == 'float64'

        # #Strings must always be treated as nominal
        # assert 'nominal' in spec['A']['flags']
    

def test__merge_spec_inheritance():
    with cre_context("test__merge_spec_inheritance") as context:
        spec1 = {"A" : "string", "B" : "number"}
        BOOP = define_fact("BOOP", spec1)

        #Should be able to inherit from ctor, type or type string
        spec2 = {"inherit_from" : BOOP, "C" : "number"}
        spec_out, inherit_from = _merge_spec_inheritance(spec2,context)
        assert inherit_from._fact_name == "BOOP"
        assert "inherit_from" not in spec_out

        spec2 = {"inherit_from" : BOOP, "C" : "number"}
        spec_out, inherit_from = _merge_spec_inheritance(spec2,context)
        assert inherit_from._fact_name == "BOOP"
        assert "inherit_from" not in spec_out

        spec2 = {"inherit_from" : "BOOP", "C" : "number"}
        spec_out, inherit_from = _merge_spec_inheritance(spec2,context)
        assert inherit_from._fact_name == "BOOP"
        assert "inherit_from" not in spec_out

        assert "A" in spec_out
        assert "B" in spec_out

        #It is illegal to redefine an attribute to have a new type
        with pytest.raises(TypeError):
            spec2 = {"inherit_from" : "BOOP", "B": "string", "C" : "string"}
            spec2 = _standardize_spec(spec2, context, "BOOP2")
            spec_out, inherit_from = _merge_spec_inheritance(spec2, context)

        #But okay to to redefine an attribute if the types match
        spec2 = {"inherit_from" : "BOOP", "B": "number", "C" : "string"}
        spec2 = _standardize_spec(spec2, context, "BOOP2")
        spec_out, inherit_from = _merge_spec_inheritance(spec2, context)


def test_define_fact():
    spec = {"A" : "string", "B" : "number"}
    spec2 = {"A" : "string", "B" : "string"}
    with cre_context("test_define_fact") as context:
        
        typ1 = define_fact("BOOP", spec)
        #Redefinition illegal with new types
        with pytest.raises(AssertionError):
            define_fact("BOOP", spec2)

    
    with cre_context("test_define_fact2") as context:
        #But is okay if defined under a different context
        typ2 = define_fact("BOOP", spec2)


def test_fact_type_pickling():
    LL_spec = {"name" : "string", "nxt" : "BOOP"}
    with cre_context("test_fact_type_pickling1") as context:
        TestLL = define_fact("TestLL", LL_spec)

    # Fact types should be picklable even if they self_reference
    typ_pickle = cloudpickle.dumps(TestLL)
    typ1 = None
    with cre_context("test_fact_type_pickling2") as context:
        # And be unpicklable
        _TestLL = cloudpickle.loads(typ_pickle)

        print(_TestLL.spec)




def test_untyped_fact():
    with cre_context("test_untyped_fact") as context:
        BOOP = define_fact("BOOP")

        @njit(cache=True)
        def make_boop():
            return BOOP(A="A", B=1.0)

        # Note: Can't do in jit context yet 
        # assert make_boop().A == "A"
        assert make_boop.py_func().A == "A"


def test_inheritence_bytes():
    assert np.array_equal(uint_to_inheritance_bytes(0xFF),[255])
    assert np.array_equal(uint_to_inheritance_bytes(0xFF00+1),[255,1])
    assert np.array_equal(uint_to_inheritance_bytes(0xFF00+0xF),[255,15])
    assert np.array_equal(uint_to_inheritance_bytes(0xFF00+0xFF),[255,255])
    assert np.array_equal(uint_to_inheritance_bytes(0xFF00+0xFF+1),[1,0,0])



def test_inheritence():
    with cre_context("test_inheritence") as context:
        spec1 = {"A" : "string", "B" : "number"}
        BOOP1 = define_fact("BOOP1", spec1)
        spec2 = {"inherit_from" : BOOP1, "C" : "number"}
        BOOP2 = define_fact("BOOP2", spec2)
        spec3 = {"inherit_from" : BOOP2, "D" : "number"}
        BOOP3 = define_fact("BOOP3", spec3)

        as_names = lambda x: [y._fact_name for y in x]

        assert as_names(context.parents_of["BOOP3"]) == ["BOOP1","BOOP2"]
        assert as_names(context.children_of["BOOP3"]) == []
        assert as_names(context.parents_of["BOOP2"]) == ["BOOP1"]
        assert as_names(context.children_of["BOOP2"]) == ["BOOP3"]
        assert as_names(context.parents_of["BOOP1"]) == []
        assert as_names(context.children_of["BOOP1"]) == ["BOOP2","BOOP3"]

        # Context should keep track of parent and child t_ids
        cd = context.context_data
        b1_t_id = BOOP1.t_id
        b2_t_id = BOOP2.t_id#cd.get_t_id(_type=BOOP2)
        b3_t_id = BOOP3.t_id#cd.get_t_id(_type=BOOP3)

        assert np.array_equal(cd.parent_t_ids[b1_t_id],[])
        assert np.array_equal(cd.parent_t_ids[b2_t_id],[b1_t_id])
        assert np.array_equal(cd.parent_t_ids[b3_t_id],[b1_t_id,b2_t_id])
        assert np.array_equal(cd.child_t_ids[b1_t_id],[b1_t_id,b2_t_id,b3_t_id])
        assert np.array_equal(cd.child_t_ids[b2_t_id],[b2_t_id,b3_t_id])
        assert np.array_equal(cd.child_t_ids[b3_t_id],[b3_t_id])

        b1 = BOOP1("A",7)
        @njit(cache=True)
        def get_idrec(b):
            return b.idrec

        assert get_idrec(b1) & 0xFF == u1(-1)
        assert get_idrec.py_func(b1) & 0xFF == u1(-1)

        b2 = BOOP2("A",7, 6)
        b3 = BOOP3("A",7, 6)

        l,p = get_inheritance_bytes_len_ptr(b1)
        print(l,p)
        l,p = get_inheritance_bytes_len_ptr(b2)
        print(l,p)
        l,p = get_inheritance_bytes_len_ptr(b3)
        print(l,p)


        t_ids1 = get_inheritance_t_ids(b1)
        t_ids2 = get_inheritance_t_ids(b2)
        t_ids3 = get_inheritance_t_ids(b3)
        print(t_ids1, t_ids2, t_ids3)

        @njit(cache=True)
        def check_isa(b1,b2,b3):
            okay = np.empty((9,),dtype=np.uint)
            okay[0] = (b1.isa(BOOP1) == 1)
            okay[1] = (b2.isa(BOOP1) == 1)
            okay[2] = (b3.isa(BOOP1) == 1)

            okay[3] = (b1.isa(BOOP2) == 0)
            okay[4] = (b2.isa(BOOP2) == 1)
            okay[5] = (b3.isa(BOOP2) == 1)

            okay[6] = (b1.isa(BOOP3) == 0)
            okay[7] = (b2.isa(BOOP3) == 0)
            okay[8] = (b3.isa(BOOP3) == 1)
            return okay

        py_okay = check_isa.py_func(b1,b2,b3)
        assert all(py_okay), str(py_okay)

        nb_okay = check_isa(b1,b2,b3)
        assert all(nb_okay), str(nb_okay)
        
def test_context_helpers():
    with cre_context("test_context_helpers") as context:
        spec1 = {"A" : "string", "B" : "number"}
        BOOP1 = define_fact("BOOP1", spec1)
        spec2 = {"inherit_from" : BOOP1, "C" : "number"}
        BOOP2 = define_fact("BOOP2", spec2)
        spec3 = {"inherit_from" : BOOP2, "D" : "number"}
        BOOP3 = define_fact("BOOP3", spec3)

        ### Check get_t_id ###
        b1_t_id = context.get_t_id(name="BOOP1")
        b2_t_id = context.get_t_id(name="BOOP2")
        b3_t_id = context.get_t_id(name="BOOP3")

        assert  context.get_t_id(_type=BOOP1) == b1_t_id 
        assert  context.get_t_id(_type=BOOP2) == b2_t_id 
        assert  context.get_t_id(_type=BOOP3) == b3_t_id 

        # assert context.get_t_id(fact_num=BOOP1._fact_num) == b1_t_id 
        # assert context.get_t_id(fact_num=BOOP2._fact_num) == b2_t_id 
        # assert context.get_t_id(fact_num=BOOP3._fact_num) == b3_t_id 

        with pytest.raises(ValueError):
            context.get_t_id()

        with pytest.raises(ValueError):
            context.get_t_id(name="SHLOOP")

        ### Check get_fact_num ###
        # b1_fact_num = context.get_fact_num(_type=BOOP1)
        # b2_fact_num = context.get_fact_num(_type=BOOP2)
        # b3_fact_num = context.get_fact_num(_type=BOOP3)

        # assert context.get_fact_num(name="BOOP1") == b1_fact_num 
        # assert context.get_fact_num(name="BOOP2") == b2_fact_num 
        # assert context.get_fact_num(name="BOOP3") == b3_fact_num 

        # assert context.get_fact_num(t_id=b1_t_id) == b1_fact_num 
        # assert context.get_fact_num(t_id=b2_t_id) == b2_fact_num 
        # assert context.get_fact_num(t_id=b3_t_id) == b3_fact_num 

        # with pytest.raises(ValueError):
        #     context.get_fact_num()

        # # with pytest.raises(ValueError):
        #     context.get_fact_num(name="SHLOOP")

        ### Check get_type ###
        # assert context.get_type(fact_num=b1_fact_num) == BOOP1 
        # assert context.get_type(fact_num=b2_fact_num) == BOOP2 
        # assert context.get_type(fact_num=b3_fact_num) == BOOP3 

        assert context.get_type(name="BOOP1") == BOOP1 
        assert context.get_type(name="BOOP2") == BOOP2 
        assert context.get_type(name="BOOP3") == BOOP3 

        assert context.get_type(t_id=b1_t_id) == BOOP1  
        assert context.get_type(t_id=b2_t_id) == BOOP2 
        assert context.get_type(t_id=b3_t_id) == BOOP3  

        with pytest.raises(ValueError):
            context.get_type()

        with pytest.raises(ValueError):
            context.get_type(name="SHLOOP")


def test_context_retroactive_register():
    with cre_context("test_context_retroactive_register") as context:
        spec1 = {"A" : "string", "B" : "number"}
        BOOP1 = define_fact("BOOP1", spec1)
        spec2 = {"inherit_from" : BOOP1, "C" : "number"}
        BOOP2 = define_fact("BOOP2", spec2)
        spec3 = {"inherit_from" : BOOP2, "D" : "number"}
        BOOP3 = define_fact("BOOP3", spec3)

    # Make a new context to ensure that we can retroactively define 
    with cre_context("other_context1") as context:
        with pytest.raises(ValueError):
            context.get_t_id(name="BOOP1")

        b1_t_id = context.get_t_id(_type=BOOP1)
        b2_t_id = context.get_t_id(_type=BOOP2)
        b3_t_id = context.get_t_id(_type=BOOP3)

        assert b1_t_id != b2_t_id and b2_t_id != b3_t_id

        c = context
        assert np.array_equal(c.get_parent_t_ids(t_id=b3_t_id),[b1_t_id,b2_t_id])
        assert np.array_equal(c.get_parent_t_ids(t_id=b2_t_id),[b1_t_id])
        assert np.array_equal(c.get_parent_t_ids(t_id=b1_t_id),[])
        
        assert np.array_equal(c.get_child_t_ids(t_id=b3_t_id),[b3_t_id])
        assert np.array_equal(c.get_child_t_ids(t_id=b2_t_id),[b2_t_id,b3_t_id])
        assert np.array_equal(c.get_child_t_ids(t_id=b1_t_id),[b1_t_id,b2_t_id,b3_t_id])

        # cd = context.context_data
        # assert np.array_equal(cd.parent_t_ids[b1_t_id],[])
        # assert np.array_equal(cd.parent_t_ids[b2_t_id],[b1_t_id])
        # assert np.array_equal(cd.parent_t_ids[b3_t_id],[b1_t_id,b2_t_id])
        # assert np.array_equal(cd.child_t_ids[b1_t_id],[b1_t_id,b2_t_id,b3_t_id])
        # assert np.array_equal(cd.child_t_ids[b2_t_id],[b2_t_id,b3_t_id])
        # assert np.array_equal(cd.child_t_ids[b3_t_id],[b3_t_id])

    




def test_cast_fact():
    with cre_context("test_cast_fact") as context:
        spec1 = {"A" : "string", "B" : "number"}
        BOOP1 = define_fact("BOOP1", spec1)
        spec2 = {"inherit_from" : BOOP1, "C" : "number"}
        BOOP2 = define_fact("BOOP2", spec2)
        spec3 = {"inherit_from" : BOOP2, "D" : "number"}
        BOOP3 = define_fact("BOOP3", spec3)

        print(BOOP1, type(BOOP1))
        print(BOOP2, type(BOOP2))
        print(BOOP3, type(BOOP3))


        b1 = BOOP1("A",7)
        b3 = BOOP3("A",1,2,3)
        bs = BaseFact()

        print(b1, type(b1))

        #upcast
        @njit
        def down_cast(b):
            # return cast_fact(BOOP1,b)    
            return b.asa(BOOP1)

        # Note w/ auto type resolution this isn't all that useful
        _b1 = down_cast(b3)
        # assert type(b1) == type(_b1)
        _b1 = down_cast.py_func(b3)    
        # assert type(b1) == type(_b1)

        #Upcast back
        @njit
        def up_cast(b):
            return b.asa(BOOP3)
            # return cast_fact(BOOP3,b)    
        _b3 = up_cast(_b1)
        assert type(b3) == type(_b3)
        _b3 = up_cast.py_func(_b1)    
        assert type(b3) == type(_b3)    

        
        #Bad cast
        FLOOP = define_fact("FLOOP", {"A" : "number", "B" : "number"})
        @njit
        def bad_cast(b):
            # return cast_fact(FLOOP,b) 
            return b.asa(FLOOP) 

        with pytest.raises(TypeError):
            bad_cast(b3)

        with pytest.raises(TypeError):
            bad_cast.py_func(b3)

        #Always allow casting to and from BaseFact
        @njit
        def base_down_cast(b):
            # return cast_fact(BaseFact,b)    
            return b.asa(BaseFact)
        _bs = base_down_cast(_b1)
        # assert type(bs) == type(_bs)
        _bs = base_down_cast.py_func(_b1)    
        # assert type(bs) == type(_bs)    

        @njit
        def base_up_cast(b):
            # return cast_fact(BOOP1,b)    
            return b.asa(BOOP1)#cast_fact(BOOP1,b)    
        _b1 = base_up_cast(_bs)
        # assert type(b1) == type(_b1)
        _b1 = base_up_cast.py_func(_bs)    
        # assert type(b1) == type(_b1)     

def test_fact_eq():
    with cre_context("test_fact_eq") as context:
        spec1 = {"A" : "string", "B" : "number"}
        BOOP1 = define_fact("BOOP1", spec1)

        b1 = BOOP1("A",7)
        b2 = BOOP1("A",7)
        b3 = BOOP1("B",8)

        @njit(cache=True)
        def do_eq(a,b):
            return a == b

        assert do_eq(b1,b2) == True
        assert do_eq(b1,b1) == True
        assert do_eq(b2,b3) == False
        assert do_eq.py_func(b1,b2) == True
        assert do_eq.py_func(b1,b1) == True
        assert do_eq.py_func(b2,b3) == False

        assert do_eq(b1,None) == False
        assert do_eq.py_func(b1,None) == False

from cre.fact_intrinsics import fact_lower_getattr

def test_getattr():
    with cre_context("test_getattr"):
        spec = {"A" : "string", "B" : "number"}
        BOOP = define_fact("BOOP", spec)

        @njit(cache=True)
        def get_it(b):
            return (b.A,b.B)

        b = BOOP("A",1)

        assert get_it.py_func(b) == ("A",1)
        assert get_it(b) == ("A",1)

        @njit(cache=True)
        def get_it_intrinsic(b):
            return (fact_lower_getattr(b,"A"),fact_lower_getattr(b,"B"))    

        # assert get_it_intrinsic.py_func(b) == ("A",1)
        assert get_it_intrinsic(b) == ("A",1)




def test_protected_mutability():
    with cre_context("test_protected_mutability") as context:
        spec = {"A" : "string", "B" : "number"}
        BOOP = define_fact("BOOP", spec)
        ms = MemSet()
        b1 = BOOP("A",0)
        b2 = BOOP("B",0)
        @njit
        def edit_it(b):
            b.B += 1


        edit_it(b1)
        edit_it(b2)
        # edit_it.py_func(b2)
        @njit
        def declare_it(ms,b,name):
            ms.declare(b,name)

        declare_it(ms,b1,"b1")
        declare_it.py_func(ms,b2,"b2")


        with pytest.raises(AttributeError):
            edit_it.py_func(b1)

        with pytest.raises(AttributeError):
            edit_it.py_func(b2)

        with pytest.raises(AttributeError):
            edit_it(b1)

        with pytest.raises(AttributeError):
            edit_it(b1)

from cre.utils import lower_setattr, PrintElapse
from cre.fact_intrinsics import fact_lower_setattr
from numba import literally


from cre.var import Var
def test_as_conditions():
    with cre_context("test_as_conditions"):
        spec = {"name" : "string", "prev" : "TestDLL", "next" : "TestDLL"}
        TestDLL = define_fact("TestDLL", spec)
        a = TestDLL(name="A")
        b = TestDLL(name="B")
        c = TestDLL(name="C")
        a.next = b
        b.prev = a
        b.next = c
        c.prev = b

        sel = Var(TestDLL, "sel")
        fact_ptr_to_var_map = {a.get_ptr() : sel, b.get_ptr(): sel.next}
        conds = a.as_conditions(fact_ptr_to_var_map)
        print(conds)

        c_ref = (sel.name == 'A') & (sel.prev == None) & (sel == sel.next.prev)
        assert str(conds) == str(c_ref)


        spec = {"name" : "string", "parent" : "TestContainer"}
        TestChild = define_fact("TestChild", spec)

        c1 = TestChild(name="c1")
        c2 = TestChild(name="c2")

        spec = {"name" : "string", "children" : "ListType(TestChild)"}
        TestContainer = define_fact("TestContainer", spec)

        c1 = TestChild(name="c1")
        c2 = TestChild(name="c2")

        C = TestContainer(name="C", children=List([c1,c2]))
        c1.parent = C
        c2.parent = C


        sel = Var(TestChild, "sel")
        fact_ptr_to_var_map = {c1.get_ptr() : sel, C.get_ptr(): sel.parent}

        conds = c1.as_conditions(fact_ptr_to_var_map)
        conds = c1.as_conditions(fact_ptr_to_var_map)
        conds = c1.as_conditions(fact_ptr_to_var_map)
        conds = c1.as_conditions(fact_ptr_to_var_map)
        conds = c1.as_conditions(fact_ptr_to_var_map)
        conds = c1.as_conditions(fact_ptr_to_var_map)
        conds = c1.as_conditions(fact_ptr_to_var_map)
        conds = c1.as_conditions(fact_ptr_to_var_map)
        conds = c1.as_conditions(fact_ptr_to_var_map)

        c_ref = (sel.name == 'c1') & (sel == sel.parent.children[0])
        assert str(conds) == str(c_ref)


# @njit
# def set_it(self,attr,other):
#     fact_lower_setattr(self,literally(attr),other)
    # self.other = other

from cre.utils import _list_base_from_ptr, _list_base, _cast_list


@njit(cache=True)
def get_base(x):
    return _list_base(x)

def test_list_type():
    with cre_context("test_list_type"):
        spec = {"A" : "string", "B" : "number"}
        BOOP = define_fact("BOOP", spec)

        spec = {"name":"string","items" : "ListType(BOOP)","other" : "BOOP"}
        BOOPList = define_fact("BOOPList", spec)

        blst_t = ListType(BaseFact)

        @njit(cache=True)
        def get_items(srl):
            items = srl.items
            if(items is not None):
                # print(_cast_list(blst_t, items))
                # print("base1",_list_base(items))    
                return srl.items
            return None

        @njit(cache=True)
        def len_items(srl):
            if(srl.items is not None):
                # no = srl.items
                # print("base2",_list_base(no))    
                return len(srl.items)
            return None

        @njit(cache=True)
        def iter_items(srl):
            if(srl.items is not None):
                # no = srl.items
                # print("base3",_list_base(no))    
                i = 0
                for p in srl.items:
                    i += 1
                return i
            return None


        a = BOOP("A",0)
        b = BOOP("B",1)
        c = BOOP("C",2)

        the_list = List([a,b])
        print("BASE!", get_base(the_list))
        bl = BOOPList("L",the_list,c)

        print("BASE!", get_base(bl.items))


        # assert str(bl) == "BOOPList(items=List([BOOP(A='A', B=0.0), BOOP(A='B', B=1.0)]))"
        # bl.other = a
        # bl.name = "BOB"
        # `
        # bl.name = "BOB"
        # set_it(bl,"other",a)
        # set_it(bl,"name","BOB")

        print(bl.idrec)
        print(bl.other)
        print(bl.items)
        print("BASE!", get_base(bl.items))

        print("START!")

        assert len(get_items(bl))  == 2
        assert len_items(bl)  == 2
        assert iter_items(bl) == 2

        print("END")

        # raise ValueError()
        # print(bl.items)

        spec = {"items" : "ListType(SelfRefList)"}
        SelfRefList = define_fact("SelfRefList", spec)

        # i =0
        # for i in range(10000):
        #     str(i)
        #     i+=1
        a = SelfRefList()
        assert len(a.items) == 0

        b = SelfRefList()
        c = SelfRefList()
        p1 = SelfRefList(List([a,b,c]))
        assert len(p1.items) == 3
        
        p1 = SelfRefList()
        p1.items = List([a,b,c])
        assert len(p1.items) == 3
        # print(p1.items)

        

        assert len(get_items(a)) == 0
        assert len(get_items(p1))  == 3
        assert len_items(p1)  == 3
        assert iter_items(p1) == 3

        # print(get_items(p1))
        # print(len_items(p1))
        # print(iter_items(p1))







        
#Work in progress...
def _test_reference_type():
    with cre_context("test_reference_type"):
        spec = {"A" : "string", "B" : "number"}
        BOOP = define_fact("BOOP", spec)

        spec = {"name" : "string", "other" : "BOOP"}
        TestRef = define_fact("TestRef", spec)

        a = BOOP("A", 1)
        b = TestRef("B", a)

        # assert str(a) == 
        # print(a, b)
        # print(a)

        # print(TestRef("B"))


        spec = {"name" : "string", "next" : "TestLL"}
        TestLL = define_fact("TestLL", spec)

        # print(TestLLType.name)

        t1 = TestLL("A")
        t2 = TestLL("B",next=t1)

        # print(t2.next)
        # print(t1,t2)



        

@njit(cache=True)
def as_cre_obj(x):
    return cast(x, CREObjType)


@njit(cache=True)
def _eq(a,b):
    return a == b

def test_eq():
    with cre_context("test_list_type"):
        spec = {"A" : "string", "B" : "number"}
        BOOP = define_fact("BOOP", spec)

        # TODO: should be range(2) waiting for numba PR #8241
        for i in range(2):
            if(i == 0): eq = _eq
            elif(i == 1): eq = operator.eq
            a1 = BOOP("HI",2)
            a2 = BOOP("HI",2)
            b1 = BOOP("HI",3)
            b2 = BOOP("HO",2)

        # print(eq(a1, a2))
            assert eq(a1, a2)
            assert not eq(a1, b1)
            assert not eq(a1, b2)

            a1 = as_cre_obj(BOOP("HI",2))
            a2 = as_cre_obj(BOOP("HI",2))
            b1 = as_cre_obj(BOOP("HI",3))
            b2 = as_cre_obj(BOOP("HO",2))

        # print(eq(a1, a2))
            assert eq(a1, a2)
            assert not eq(a1, b1)
            assert not eq(a1, b2)

        

@njit(cache=True)
def _hsh(x):
    return hash(x)


def test_hash():
    with cre_context("test_hash"):
        spec = {"A" : "string", "B" : "number"}
        BOOP = define_fact("BOOP", spec)

        # TODO: should be range(2) waiting for numba PR #8241
        for i in range(1):
            if(i == 0): hsh = _hsh
            elif(i == 1): hsh = hash

            a1 = BOOP("HI",2)
            a2 = BOOP("HI",2)
            b1 = BOOP("HI",3)
            b2 = BOOP("HO",2)

            assert hsh(a1) == hsh(a2)
            assert hsh(a1) != hsh(b1)
            assert hsh(a1) != hsh(b2)

            a3_boop = BOOP("HI",2)
            a3 = as_cre_obj(a3_boop)
            assert hsh(a3) == hsh(a1)

            # check that mutation causes rehash
            a3_boop.B = 7
            assert hsh(a3) != hsh(a1)


            a1 = as_cre_obj(BOOP("HI",2))
            a2 = as_cre_obj(BOOP("HI",2))
            b1 = as_cre_obj(BOOP("HI",3))
            b2 = as_cre_obj(BOOP("HO",2))

            assert hsh(a1) == hsh(a2)
            assert hsh(a1) != hsh(b1)
            assert hsh(a1) != hsh(b2)

            a3_boop = BOOP("HI",2)
            a3 = as_cre_obj(a3_boop)
            assert hsh(a3) == hsh(a1)

            # check that mutation causes rehash
            a3_boop.B = 7
            assert hsh(a3) != hsh(a1)



def test_copy():
    with cre_context("test_copy"):
        spec = {"A" : "string", "B" : "number"}
        BOOP = define_fact("BOOP", spec)
        MOOP = define_fact("MOOP", {'boop1' : BOOP, 'boop2' :BOOP})

        
        b = BOOP("1",2)
        a = copy_cre_obj(b)

            
        @njit(cache=True)
        def do_primitive_copy():
            a1 = BOOP("HI",2)
            a2 = copy_cre_obj(a1)
            a1.A = "NOT HI"
            a2.B = 0
            return a1, a2

        a1, a2 = do_primitive_copy()
        assert a2.A == "HI" and a2.B == 0
        assert a1.A != a2.A and a1.B != a2.B

        a1, a2 = do_primitive_copy.py_func()
        assert a2.A == "HI" and a2.B == 0
        assert a1.A != a2.A and a1.B != a2.B

        @njit(cache=True)
        def do_obj_copy():
            b1 = BOOP("HI",2)
            b2 = BOOP("HO",3)
            m1 = MOOP(b1,b2)
            m2 = copy_cre_obj(m1)
            m2.boop2 = BOOP("HE",4)
            b1,b2 = None, None
            return m1, m2

        m1, m2 = do_obj_copy()
        assert m2.boop1.B == 2
        assert m1.boop2.B == 3
        assert m2.boop2.B == 4

        m1, m2 = do_obj_copy.py_func()
        assert m2.boop1.B == 2
        assert m1.boop2.B == 3
        assert m2.boop2.B == 4

def test_repr():
    with cre_context("test_repr"):
        BOOP = define_fact("BOOP",{"A": "string", "B" : "number", "C" : "BOOP"})
        c = BOOP("C",7)


        og_refcount = c._meminfo.refcount 
        b = BOOP("A", 1, c)

        assert "BOOP(A='A', B=1.0, C=<BOOP at " in repr(b)

        # Call str on b several times + make sure no memleak in c
        [str(b) for i in range(10)]
        print(og_refcount, c._meminfo.refcount)
        assert c._meminfo.refcount <= og_refcount + 2 # Probably should be 1 here





# def test_weird_case():
#     from cre.fact_intrinsics import fact_lower_getattr
#     from cre.memset import MemSet
#     with cre_context("test_weird_case"):
#         TestLL = define_fact("TestLL",{
#             "id": "string",
#             "value" : {"type": "string", "is_semantic_visible" : True},
#             "nxt" : "TestLL",
#         })

#         ms = MemSet()
#         q = TestLL(id="q", value="1")
#         ms.declare(q)
#         @njit(cache=True)
#         def foo(x):
#             print(fact_lower_getattr(x,"id"))
#             # return x.id
#         foo(q)



@generated_jit(cache=True)
def _b_boop_ctor_10000():
    with cre_context("_b_boop_ctor_10000"):
        # def _define_boop():
        #     with cre_context("_b_boop_ctor_1000"):
        spec = {"A" : "string", "B" : "number"}
        BOOP = define_fact("BOOP", spec)
                # return (BOOP,), {}
    def impl():
        for i in range(10000):
            b = BOOP("HI",i)
    return impl

    
    # py_b_boop_ctor_100 = py_b_boop_ctor_100.py_func

@pytest.mark.benchmark(group="fact")
def test_b_boop_ctor_10000(benchmark):
    with cre_context("_b_boop_ctor_10000"):
        benchmark.pedantic(_b_boop_ctor_10000, warmup_rounds=1, rounds=10)


# @njit(cache=True)
def _b_py_dict_boop_10000():
    for i in range(10000):
        b = {"A" : "HI", "B" : i}

@pytest.mark.benchmark(group="fact")
def test_b_py_dict_boop_10000(benchmark):
    with cre_context("_b_boop_ctor_10000"):
        benchmark.pedantic(_b_py_dict_boop_10000, warmup_rounds=1, rounds=10)



# def test_b_py_dict_boop_100(benchmark):
#     benchmark.pedantic(_b_py_dict_boop(100), warmup_rounds=1, rounds=10)




if __name__ == "__main__":
    import faulthandler; faulthandler.enable()

    # test_context_helpers()
    # test_context_retroactive_register()
    # test_getattr()
    
    # test_list_type()
    # _test_reference_type()
    # test__standardize_spec()
    # test__merge_spec_inheritance()
    # test_define_fact()
    # test_inheritence()
    # test_cast_fact()
    # test_protected_mutability()
    # test_fact_eq()

    # test_untyped_fact()
    # test_fact_eq()

    test_as_conditions()

    # _test_reference_type()
    # test_hash()
    # test_eq()
    # test_copy()
    # test_weird_case()
    # test_inheritence()
    # test_inheritence_bytes()
    # test_context_retroactive_register()

    # test_repr()





##### NOTES NOTES NOTES ####
'''
[] Need a way to organize tests -> need to have some standard types:
    -TObj = {"name" : "string", "value" : "number"}
    -TRef = {"name" : "string", "other" : "TestObj"}
    -TLL = {"name" : "string", "next" : "TLL"}
    -TList = {"name" : "string", "items" : "ListType("string")"}
    -TSelfList = {"name" : "string", "items" : "ListType("TSelfList")"}
    -TObjList = {"name" : "string", "items" : "ListType("TObj")"}

[] Should test alternative definition patterns
    -string vs type obj
    
[] For each type need to test:
    -getattr, setattr, str 
    -mem.modify
    -matching




'''



from numba import njit, i8,i4,u1, u2, u8, f8, f4, generated_jit, types
from numba.types import ListType, DictType, Tuple
from numba.typed import Dict, List
from cre.vector import VectorType
from cre.structref import define_structref
from cre.fact import BaseFact
from cre.utils import decode_idrec, encode_idrec
from numba.experimental.structref import new

RETRACT = u1(0xFF)# u1(0)
DECLARE = u1(0)

change_event_fields = {
    # The idrec for the declared/modified/retracted object
    "idrec" : u8,
    # The t_id for the declared/modified/retracted object
    "t_id" : u2,
    # The f_id for the declared/modified/retracted object
    "f_id" : u8,
    
    # Whether or not a Fact with this idrec was ever retracted 
    #  indicating that we might need to clean something up
    "was_retracted" : u1,

    # Whether or not a Fact with this idrec was declared. If
    #  both a retract and a declare occured then 'was_retracted'
    #  will be true, but 'was_declared' will only be true if it
    #  occured after the retraction.
    "was_declared" : u1,

    # Whether or not a Fact with this idrec was modified. Will be False 
    #  if a declare or retract occured.
    "was_modified" : u1,

    # If a modification occured the a_ids associated with the change
    "a_ids" : types.optional(ListType(u1)),

}

ChangeEvent, ChangeEventType = define_structref("ChangeEvent", change_event_fields, define_constructor=False) 
ChangeEvent.__str__ = lambda x : f"ChangeEvent(t_id={x.t_id}, f_id={x.f_id}, ret={x.was_retracted}, dec={x.was_declared}, mod={x.was_modified}{f', a_ids={x.a_ids}' if x.a_ids else ''})"

@njit(cache=True)
def change_event_ctor(idrec):
    st = new(ChangeEventType)
    st.idrec = idrec
    st.t_id, st.f_id, _ = decode_idrec(idrec)
    st.a_ids = None
    st.was_retracted = False
    st.was_declared = False
    st.was_modified = False
    return st    


# change_event_ctor(0)
u1_opBaseFact_Tuple_type = Tuple((u8,u1[::1],types.optional(BaseFact)))
@njit(ListType(ChangeEventType)(VectorType,i8,i8), cache=True, locals={"a_id" : u1})
def accumulate_change_events(cq, start, end=-1):
    ''' Takes in a change queue (i.e. Vector<i8>) and a start and end position
        and returns an accumulated set of ChangeEvent objects associated with
        the changes between start and end. If end is not set then it defaults
        to the head of the the change_queue.
    '''
    # cq = incr_pr.in_mem.mem_data.change_queue
    if(end == -1): end = cq.head
    if(start == end):
        return List.empty_list(ChangeEventType)

    ce_dict = Dict.empty(u8,ChangeEventType)
    for i in range(start, end):
        t_id, f_id, a_id = decode_idrec(cq[i])
        # print(t_id, f_id, a_id)
        idrec = encode_idrec(t_id, f_id, u1(0))
        if(idrec not in ce_dict):
            ce_dict[idrec] = change_event_ctor(idrec)
        ce = ce_dict[idrec]

        if(a_id == RETRACT):
            ce.was_modified = False
            ce.was_declared = False
            ce.was_retracted = True
            ce.a_ids = None
        elif(a_id == DECLARE):
            ce.was_modified = False
            ce.was_declared = True
            ce.a_ids = None
        else:
            # Modify case
            if(not ce.was_declared):
                ce.was_modified = True
                if(ce.a_ids is None):
                    ce.a_ids = List.empty_list(u1)
                if(a_id not in ce.a_ids):
                    ce.a_ids.append(a_id)
    change_events = List.empty_list(ChangeEventType)
    for ce in ce_dict.values():
        change_events.append(ce)
    return change_events
